-- Helperfunctions
local string_lower, string_gsub, string_format = string.lower, string.gsub, string.format
local _, wipe, tsort, tinsert = _, wipe, table.sort, table.insert
local floor, hugeNumber, max, min = math.floor, math.huge, math.max, math.min
local GetRemainingCompletions = LFGRewardsFrame_EstimateRemainingCompletions

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local Ailo = LibStub("AceAddon-3.0"):NewAddon("Ailo", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Ailo", false)
local LST = LibStub("LibSyncTime-1.0", true)
local DB_VERSION = 10
local currentChar, currentRealm, currentCharRealm, currentMaxLevel
local icons = setmetatable( {}, { __index = function() return "Interface\\Icons\\INV_Misc_QuestionMark" end } )

function Ailo:Output(...)
    if self.db.profile.showMessages then
        Ailo:Print(...)
    end
end 

-- Sorting
local sortRealms = {}
local sortRealmsPlayer = {}

local RAID_CLASS_COLORS_FONTS = {}

local function setColor(info, r, g, b, a)
    Ailo.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
end

local function getColor(info)
    return Ailo.db.profile[info[#info]].r, Ailo.db.profile[info[#info]].g, Ailo.db.profile[info[#info]].b, Ailo.db.profile[info[#info]].a
end

local function GetWeeklyResetDay()
    -- number is index into day of week; 1 == Sunday
    local resets = { 
        enCN = 3, enUS = 3, esMX = 3, ptBR = 3,                                   -- Tuesday 
        deDE = 4, enGB = 4, esES = 4, frFR = 4, itIT = 4, ptPT = 4, ruRU = 4,     -- Wednesday
        enTW = 5, koKR = 5, zhCN = 5, zhTW = 5,                                   -- Thursday
    }

    local locale = GetLocale()
    local reset = resets[locale]

    if not reset then Ailo:Print(locale, "locale is not recognized, please report this error") end
    return reset or 3
end

-- updateRetry variables used by FirstUpdate timer callback
local updateRetryLimit = 10
local updateRetryCount = 0
local updateRetryTimer = nil

local function GetNextDailyReset()
    -- when first logging in, we may get a bogus result from GetQuestResetTime
    -- if thats the case, then just return hugeNumber so that nothing gets purged
    local qrtime = GetQuestResetTime()
    if qrtime <= 0 then
        return hugeNumber
    end

    return Ailo:GetSyncTime() + qrtime
end

local function GetNextWeeklyReset()
    -- if we're not getting a valid number for the daily reset, then we can't
    -- calculate the weekly. just return hugeNumber so that nothing gets purged 
    local nextDailyReset = GetNextDailyReset()
    if (nextDailyReset == hugeNumber) then
        return hugeNumber
    end

    local dayOfNextReset = date("*t", nextDailyReset).wday
    local daysToWeeklyReset =  math.fmod(7 + Ailo.db.global.weeklyResetDay - dayOfNextReset, 7)
    local nextWeeklyReset = nextDailyReset + (daysToWeeklyReset * 86400) -- 86400 is number of seconds in a day (24 * 60 * 60)

    return nextWeeklyReset
end


--=========================================================================--
-- data table definitions
--=========================================================================--

local defaults = {
    profile = {
        savedraid = { r=1, g=0, b=0, a=1 },
        freeraid  = { r=0, g=1, b=0, a=1 },
        hc  = "hc",
        nhc = "nhc",
        show5Man             = false,
        showAllChars         = false,
        showCharacterRealm   = false,
        showWeeklyValor      = true,
        showDailyHeroic      = true,
        showDailyScenario    = false,
        showLeft             = false,
        showMessages         = true,
        showRealmHeaderLines = false,
        showRaidFinder       = true,
        showTBVictory        = false,
        showWGVictory        = false,
        showWeeklyConquest   = false,
        showDailyPVP         = false,
        showSeasonal         = true,
        showShaOfAnger       = false,
        useClassColors       = true,
        useCustomClassColors = true,
        instanceAbbr = {},
        minimapIcon = {
            hide = false,
            minimapPos = 220,
            radius = 80,
        },
        
    },
    global = {
        chars = {},             -- structure is = { realm = { char = { lockouts = { expire, encounter }, resets = { expire, encounter } } } }  
        charClass = {},
        raids = {},             -- structure is = { instance = { size } }
        lfrs = {},              -- structure is = { id = expire }
        nextPurge = hugeNumber,
        weeklyResetDay = 3,     -- TODO - set default based on region
        version = 0,
    },
}

local Seasonal = {}
Seasonal.Events = {
    Valentines = { icon = "Interface\\Icons\\inv_valentinesboxofchocolates02", 
                  first = time( { year = 2013, month = 2,  day = 10, hour =  0, min =  1 } ),
                   last = time( { year = 2013, month = 2,  day = 23, hour = 23, min = 59 } ),
                  level = 1,
             dungeon_id = 288 },
     Midsummer = { icon = "Interface\\Icons\\inv_misc_bag_17", 
                  first = time( { year = 2013, month = 6,  day = 21, hour =  0, min =  1 } ),
                   last = time( { year = 2013, month = 7,  day =  4, hour = 23, min = 59 } ),
                  level = 1,
             dungeon_id = 286 },
      Brewfest = { icon = "Interface\\Icons\\inv_cask_02", 
                  first = time( { year = 2013, month = 9,  day = 20, hour =  0, min =  1 } ),
                   last = time( { year = 2013, month = 10, day =  5, hour = 23, min = 59 } ),
                  level = 1,
             dungeon_id = 287 },
    HallowsEnd = { icon = "Interface\\Icons\\inv_misc_bag_28_halloween", 
                  first = time( { year = 2013, month = 10, day = 18, hour =  1, min =  0 } ),
                   last = time( { year = 2013, month = 10, day = 31, hour = 23, min = 59 } ),
                  level = 1,
             dungeon_id = 285 },
    WinterVeil = { icon = "Interface\\Icons\\inv_holiday_christmas_present_01",
                  first = time( { year = 2012, month = 12, day = 15, hour = 6 } ),
                   last = time( { year = 2013, month = 1,  day = 2,  hour = 6 } ),
                  level = 0,
              quest_ids = { 6983, 7043 }, },
}

local RaidFinder = { 
    { id = 613, count = 3, GetExpireTime = GetNextWeeklyReset }, -- Pinnacle of Storms
    { id = 612, count = 3, GetExpireTime = GetNextWeeklyReset }, -- Halls of Flesh-Shaping
    { id = 611, count = 3, GetExpireTime = GetNextWeeklyReset }, -- Forgotten Depths
    { id = 610, count = 3, GetExpireTime = GetNextWeeklyReset }, -- Last Stand of the Zandalari
    { id = 526, count = 4, GetExpireTime = GetNextWeeklyReset }, -- Terrace of Endless Spring
    { id = 529, count = 3, GetExpireTime = GetNextWeeklyReset }, -- The Dread Approach
    { id = 530, count = 3, GetExpireTime = GetNextWeeklyReset }, -- Nightmare of Shek'zeer
    { id = 527, count = 3, GetExpireTime = GetNextWeeklyReset }, -- Guardians of Mogu'shan
    { id = 528, count = 3, GetExpireTime = GetNextWeeklyReset }, -- The Vault of Mysteries
    { id = 416, count = 4, GetExpireTime = GetNextWeeklyReset }, -- The Siege of Wyrmrest Temple
    { id = 417, count = 4, GetExpireTime = GetNextWeeklyReset }, -- Fall of Deathwing
}

local RepeatableQuests = {
    -- ID's of horde and alliance versions of Victory in Wintergrasp weekly pvp quest 
    wgvictory = { GetExpireTime = GetNextWeeklyReset, questIds = { 13181, 13183 } },
    -- ID's of horde and alliance versions of the Victory in Tol Barad weekly quest 
    tbvictory = { GetExpireTime = GetNextWeeklyReset, questIds = { 28882, 28884 } },
    -- ID's of all horde and alliance versions of repeatable Seasonal quests
    -- Note: supports only one h/a set of quests per event, but all events can live
    -- in same struct as long as they don't have overlapping run times.
    seasonal = { GetExpireTime = GetNextDailyReset, questIds = { 6983, 7043 } },
    -- asssuming quest id is same for horde and alliance
    shaofanger = { GetExpireTime = GetNextWeeklyReset, questIds = { 32099 } },
}

local Points = {
    weeklyvalor    = { id = _G.VALOR_CURRENCY,    total = -1, scaler = .01 },
    weeklyconquest = { id = _G.CONQUEST_CURRENCY, total = -1, scaler = 1   },
}


--=========================================================================--
--=========================================================================--
--
-- Addon Initialization
--
--=========================================================================--
--=========================================================================--

function Ailo:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
    defaults.global.weeklyResetDay = GetWeeklyResetDay()
    self.db = LibStub("AceDB-3.0"):New("AiloDB", defaults, true)
    self:MigrateSavedData()

    LibStub("AceConfig-3.0"):RegisterOptionsTable("Ailo", self.GenerateOptions)
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Ailo")

    local AiloLDB = ldb:NewDataObject("Ailo", {
        type = "data source",
        text = "Ailo",
        icon = "Interface\\Icons\\Achievement_Dungeon_UlduarRaid_Archway_01.png",
        OnClick = function(clickedframe, button)
            if button == "RightButton" then 
                InterfaceOptionsFrame_OpenToCategory(Ailo.optionsFrame) 
            else 
                if IsShiftKeyDown() then
                    Ailo:ManualPlayerUpdate() 
                else
                    ToggleFriendsFrame(4)
                end
            end
        end,
        OnEnter = function(tt)
            local tooltip = LibQTip:Acquire("AiloTooltip", 1, "LEFT") 
            Ailo.tooltip = tooltip
            Ailo:PrepareTooltip(tooltip) 
            tooltip:SmartAnchorTo(tt)
            tooltip:Show()
        end,
        OnLeave = function(tt)
            LibQTip:Release(Ailo.tooltip)
            Ailo.tooltip = nil
        end,
    })

    LDBIcon:Register("Ailo", AiloLDB, self.db.profile.minimapIcon)
end

function Ailo:OnEnable() -- PLAYER_LOGIN, game data is initialized
    currentMaxLevel = MAX_PLAYER_LEVEL_TABLE[GetExpansionLevel()] - Seasonal:LevelAdjustment()
    currentChar = UnitName("player")
    currentRealm = GetRealmName()
    currentCharRealm = currentChar..' - '..currentRealm

    self:InitIconsTable()

    self:SetupClasscoloredFonts()
    if CUSTOM_CLASS_COLORS then
        CUSTOM_CLASS_COLORS:RegisterCallback("SetupClasscoloredFonts", self)
    end

    if currentMaxLevel <= UnitLevel("player") then 
        updateRetryTimer = self:ScheduleRepeatingTimer("DelayedInit", .2)
    elseif self.db.global.chars[currentRealm] then
        -- This character was once elligible to be shown and so would still be shown if 
        -- showAllChars is true. Clear them out chars to make sure no row shows for them.
        self.db.global.chars[currentRealm][currentChar] = nil
        self:TrimRaidTable()
    end
end

function Ailo:DelayedInit()
    -- When GetNextDailyReset is called early, like during OnEnable, it can 
    -- return hugeNumber. If a hugeNumber is saved as an expire time acedb
    -- will convert it to a nil. This is most often a problem for currency
    -- trackers which could only check once, during the first UpdatePlayer
    -- called during OnEnable. Solution is to use an Ace Timer callback to
    -- introduce a small delay between OnEnable and PlayerUpdate to give 
    -- the game time get the daily quest reset time sorted out.
    -- If we hit RetryLimit, stop delaying and just let it init.

    if updateRetryCount < updateRetryLimit then
        updateRetryCount = updateRetryCount + 1
        if GetNextDailyReset() == hugeNumber then
            return
        end
    end

    if updateRetryTimer then
        self:CancelTimer(updateRetryTimer, true)
        updateRetryCount = updateRetryLimit
        updateRetryTimer = nil
    end

    self:RegisterEvent("CHAT_MSG_SYSTEM")
    self:RegisterEvent("UPDATE_INSTANCE_INFO", "UpdateLockouts")
    self:RegisterEvent("LFG_COMPLETION_REWARD")
    self:RegisterEvent("LFG_UPDATE_RANDOM_INFO", "UpdateLFGResets")
    self:RegisterEvent("CURRENCY_DISPLAY_UPDATE", "UpdatePoints")
    self:RegisterEvent("UNIT_QUEST_LOG_CHANGED")
    self:RegisterEvent("PLAYER_REGEN_ENABLED", "UpdateQuests")

    -- trigger collection of lockout, reset and quest information
    self:UpdatePlayer()
end   

function Ailo:OnDisable()
    self:CancelAllTimers()
end

-- if we change the structure or contents of the saved settings, we can 
-- do a fixup of the data here the first time we run after the upgrade
function Ailo:MigrateSavedData()
    local userVersion = self.db.global.version

    -- if there is no version stamp, only safe thing to do is wipe it
    if not userVersion then
        Ailo:WipeDB()
        self.db.global.version = DB_VERSION 
    -- if version is 0 then its a new user and so there's nothing to update but the version number
    elseif userVersion == 0 then
        self.db.global.version = DB_VERSION
    -- if version is less current, settings may need to be migrated/tweaked/etc
    elseif userVersion < DB_VERSION then
        if userVersion < 9 then
            -- change to values stored for daily heroic 
            self:Output(L["DB_VERSION_UPGRADE_PURGE"])
            Ailo:WipeDB()

            -- if they weren't showing the old VP column, don't show the new
            self.db.profile.showWeeklyValor = self.db.profile.showDailyHeroic
        end

        if userVersion and userVersion < 10 then
            -- database does not need to be wiped for these changes

            -- if they were showing the daily PVP column, they'll likely want weekly CP tracking
            self.db.profile.showWeeklyConquest = self.db.profile.showDailyPVP
            -- remove dalaran weekly raid quest tracking
            showWeeklyRaid = nil
        end

        self.db.global.version = DB_VERSION 
    end
end

function Ailo:InitIconsTable()
    icons.Valor    = select(3, GetCurrencyInfo(_G.VALOR_CURRENCY))
    icons.Justice  = select(3, GetCurrencyInfo(_G.JUSTICE_CURRENCY))
    icons.Conquest = select(3, GetCurrencyInfo(_G.CONQUEST_CURRENCY))
    icons.Honor    = select(3, GetCurrencyInfo(_G.HONOR_CURRENCY))

    icons.RandomBG = "Interface\\PVPFrame\\Icon-Combat"
    icons.LFR      = "Interface\\LFGFrame\\UI-LFR-PORTRAIT"
    icons.LFD      = "Interface\\Icons\\INV_Helmet_08"
    icons.LFS      = "Interface\\Icons\\Icon_Scenarios"

    icons.TolBarad    = select(3, GetCurrencyInfo(391))  -- Tol Barad Commendation
    icons.Wintergrasp = select(3, GetCurrencyInfo(126))  -- Wintergrasp Mark of Honor

    icons.ShaOfAnger = GetItemIcon(90839)  -- Cache of Sha-Touched Gold
end

--=========================================================================--
--=========================================================================--
--
-- main data acquisition and processing function
--
--=========================================================================--
--=========================================================================--

function Ailo:UpdatePlayer()
    if currentMaxLevel > UnitLevel("player") then return end
    self:PrepareCharacterTables()
    self.db.global.charClass[currentCharRealm] = select(2,UnitClass('player'))

    self:UpdateQuests()
    self:UpdatePoints()

    -- update lockouts for raids and instances by triggering an UPDATE_INSTANCE_INFO event
    RequestRaidInfo()

    -- update resets and LFG rewards by triggering an LFG_UPDATE_RANDOM_INFO event 
    RequestLFDPlayerLockInfo()

    -- ensure we don't miss a reset
    self.db.global.nextPurge = self:GetNextPurge()
end

--=========================================================================--
-- Currency Point Trackers
--=========================================================================--

-- event hander for CURRENCY_DISPLAY_UPDATE
function Ailo:UpdatePoints(...)
    local resets = self.db.global.chars[currentRealm][currentChar].resets
    local expire = GetNextWeeklyReset()
    local updateresets = false

    for k, v in pairs(Points) do
        -- GetCurrencyInfo returns weeklyMax = 100000 and totalMax = 300099 for VALOR_CURRENCY
        -- the amount and earnedThisWeek values on the other hand are correct
        -- all values returned by CONQUEST_CURRENCY look fine
        -- hack work around by adding a scaler value for currencies in the points table
        local _, amount, _, earnedThisWeek, weeklyMax, totalMax = GetCurrencyInfo(v.id)
        weeklyMax = floor(v.scaler * weeklyMax)
        totalMax  = floor(v.scaler * totalMax)
        local isCapped = earnedThisWeek >= weeklyMax or amount >= totalMax

        updateresets = updateresets or v.total ~= amount
        v.total = amount
        resets[k] = { expire = expire, encounter = earnedThisWeek, done = isCapped }
    end

    if updateresets then
        RequestLFDPlayerLockInfo()
    end
end

--=========================================================================--
-- Raid and Instance Lockouts
--=========================================================================--

local INSTANCE_SAVED = _G["INSTANCE_SAVED"]
function Ailo:CHAT_MSG_SYSTEM(event, msg)
    if tostring(msg) == INSTANCE_SAVED then
        -- You are now saved to this instance; refresh Lockout info
        RequestRaidInfo()  -- triggers UPDATE_INSTANCE_INFO event
    end
end

-- event handler for UPDATE_INSTANCE_INFO, process all raid/instance lockouts
function Ailo:UpdateLockouts()
    if currentMaxLevel > UnitLevel("player") then return end
    self.db.global.charClass[currentCharRealm] = select(2,UnitClass('player'))

    local now = self:GetSyncTime()
    for index=1, GetNumSavedInstances() do
        local instanceName, _, instanceReset, _, locked, _, _, isRaid, maxPlayers, _, maxEncounter, encounterDone = GetSavedInstanceInfo(index)
        if maxPlayers == 25 then 
            maxPlayers = 10 
        end
        if locked then
            local instanceString = string_format("%s.%d", instanceName, maxPlayers)
            self:ExtendRaidTable( instanceName, maxPlayers )
            self:SaveRaidForChar( instanceString, now+instanceReset, maxEncounter-encounterDone )
        end
    end
    
end

function Ailo:SaveRaidForChar(instance, expireTime, encounterLeft, character, realm)
    character     = character or currentChar
    realm         = realm or currentRealm
    encounterLeft = encounterLeft or 0

    local lockouts = self.db.global.chars[realm][character].lockouts
    lockouts[instance] = lockouts[instance] or {}
    lockouts[instance].expire    = expireTime
    lockouts[instance].encounter = encounterLeft

    if expireTime < self.db.global.nextPurge then
        self.db.global.nextPurge = expireTime
    end
end

--=========================================================================--
-- LFG Reward Resets
--=========================================================================--

function Ailo:LFG_COMPLETION_REWARD()
    --[[
    Fires when a random dungeon is completed and the achievement-like
    alert window pops up. The problem is that this DOES NOT update
    the return values of GetLFGDungeonRewards(x), those are updated
    when LFG_UPDATE_RANDOM_INFO is recieved, so force the client to
    call for an update which will be handled by UpdateLFGResets.
    ]]--
    RequestLFDPlayerLockInfo()
end

-- event handler for LFG_UPDATE_RANDOM_INFO, process all weekly/daily instance resets
function Ailo:UpdateLFGResets(...)
    local resets = self.db.global.chars[currentRealm][currentChar].resets
    local lfrs = self.db.global.lfrs

    -- 462 random mists of pandaria heroic
    -- 463 random mists of pandaria dungeon
    -- 493 random mists of pandaria scenario

    -- cell is red only if instance type is heroic and if done today or has no additional rewards (means money only, no valor)
    local bestchoice = GetRandomDungeonBestChoice() or 462
    if select(LFG_RETURN_VALUES.difficulty, GetLFGDungeonInfo(bestchoice)) > 0 then
        local doneToday, _, _, _, _, numRewards = GetLFGDungeonRewards(462)
        if numRewards < 1 then  -- if not returning any valor then we must be capped for the week
            resets.dailyheroic = { expire = GetNextWeeklyReset() }
        elseif doneToday then
            resets.dailyheroic = { expire = GetNextDailyReset() }
        else -- if they hit total cap we mark as done, need to clear that if they spend points and drop under caps again
            resets.dailyheroic = nil
        end
    end

    -- daily scenario first completion bonus
    local doneToday, _, _, _, _, numRewards = GetLFGDungeonRewards(493)
    if numRewards < 2 then  -- if not returning any valor then we must be capped for the week
        resets.dailyscenario = { expire = GetNextWeeklyReset() }
    elseif doneToday then
        resets.dailyscenario = { expire = GetNextDailyReset() }
    else -- if they hit total cap we mark as done, need to clear that if they spend points and drop under caps again
        resets.dailyscenario = nil
    end

    -- daily battleground first win bonus
    if GetRandomBGHonorCurrencyBonuses() then
        resets.dailypvp = { expire = GetNextDailyReset() }
    end

    local id = Seasonal:Dungeon()
    if (id and GetLFGDungeonRewards(id)) then
        resets.seasonal = { expire = GetNextDailyReset() }
    end

    for _, rf in pairs(RaidFinder) do
        local _, killed = GetLFGDungeonNumEncounters(rf.id)
        if killed > 0 then
            local expire = rf.GetExpireTime()
            resets[rf.id] = { expire = expire, encounter = (rf.count - killed), done = GetLFGDungeonRewards(rf.id) }
            lfrs[rf.id] = max(expire, lfrs[rf.id] or 0)
        end
    end
end

--=========================================================================--
-- Daily/Weekly Repeatable Quests
--=========================================================================--

function Ailo:UNIT_QUEST_LOG_CHANGED(event, unitID, ...)
    if unitID == "player" then
        self:UpdateQuests()
    end
end

function Ailo:UpdateQuests()
    local resets = self.db.global.chars[currentRealm][currentChar].resets

    for objective,data in pairs(RepeatableQuests) do
        for _,id in pairs (data.questIds) do
            if IsQuestFlaggedCompleted(id) then
                resets[objective] = { expire = data.GetExpireTime() }
                break
            end
        end
    end
end

--=========================================================================--
--=========================================================================--
--
-- saved data maintenance functions
--
--=========================================================================--
--=========================================================================--

function Ailo:PurgeExpiredData()
    local now = self:GetSyncTime()

    for realm, chars in pairs(self.db.global.chars) do 
        for char, charData in pairs(chars) do 
            if charData.lockouts then 
                for instance, instanceData in pairs(charData.lockouts) do
                    if type(instanceData) ~= "table" or not instanceData.expire or now > instanceData.expire then
                        self.db.global.chars[realm][char].lockouts[instance] = nil
                    end
                end
            end
            if charData.resets then 
                for instance, instanceData in pairs(charData.resets) do
                    if type(instanceData) ~= "table" or not instanceData.expire or now > instanceData.expire then
                        self.db.global.chars[realm][char].resets[instance] = nil
                    end
                end
            end
        end
    end

    local lfrsdb  = self.db.global.lfrs
    for id, expire in pairs(lfrsdb) do
        if now > expire then
            lfrsdb[id] = nil
        end
    end
end

function Ailo:GetNextPurge()
    local nextPurge = GetNextDailyReset()

    for realm, chars in pairs(self.db.global.chars) do 
        for _, charData in pairs(chars) do 
            if charData.lockouts then
                for _, instanceData in pairs(charData.lockouts) do
                    nextPurge = min(nextPurge, instanceData.expire)
                end
            end
            if charData.resets then
                for _, instanceData in pairs(charData.resets) do
                    nextPurge = min(nextPurge, instanceData.expire)
                end
            end
        end
    end

    for _, expire in pairs(self.db.global.lfrs) do
        nextPurge = min(nextPurge, expire)
    end

    return nextPurge
end

function Ailo:GetSyncTime()
    return LST and LST:synctime() or time()
end

function Ailo:WipeDB()
    if type(self.db.global.chars) == "table" then 
        wipe(self.db.global.chars)
    end
    if type(self.db.global.raids) == "table" then 
        wipe(self.db.global.raids)
    end
    if type(self.db.global.lfrs) == "table" then 
        wipe(self.db.global.lfrs)
    end
    if type(self.db.global.charClass) == "table" then 
        wipe(self.db.global.charClass)
    end
    self.db.global.nextPurge = GetNextDailyReset()
end

function Ailo:PrepareCharacterTables()
    if not self.db.global.chars[currentRealm] then
        self.db.global.chars[currentRealm] = {}
    end
    if not self.db.global.chars[currentRealm][currentChar] then
        self.db.global.chars[currentRealm][currentChar] = {}
    end
    if not self.db.global.chars[currentRealm][currentChar].lockouts then
        self.db.global.chars[currentRealm][currentChar].lockouts = {}
    end
    if not self.db.global.chars[currentRealm][currentChar].resets then
        self.db.global.chars[currentRealm][currentChar].resets = {}
    end
end


function Ailo:ExtendRaidTable(instanceName, size)
    -- Save it for the tooltip
    self.db.global.raids[instanceName] = self.db.global.raids[instanceName] or {}
    self.db.global.raids[instanceName][size] =  true
end

function Ailo:TrimRaidTable()
    local raidsdb = self.db.global.raids
    for raid, sizes in pairs(raidsdb) do
        for size in pairs(sizes) do
            local isUsed = false
            local raidString = string_format("%s.%d", raid, size)
            for realm, chars in pairs(self.db.global.chars) do 
                for char, charData in pairs(chars) do
                    if charData.lockouts and charData.lockouts[raidString] and charData.lockouts[raidString].expire then
                        isUsed = true
                        break
                    end
                end
            end
            if not isUsed then
                raidsdb[raid] = nil
            end
        end
    end
end


function Ailo:ManualPlayerUpdate()
    if currentMaxLevel > UnitLevel("player") then return end
    self:Output(L["Updating data for current player."])

    if not self.db.global.chars[currentRealm] then 
        self.db.global.chars[currentRealm] = {}
    else
        wipe(self.db.global.chars[currentRealm][currentChar])
        self.db.global.chars[currentRealm][currentChar] = nil
    end

    self:TrimRaidTable()
    self:UpdatePlayer()
end


function Ailo:GetInstanceAbbr(instanceName)
    if not self.db.profile.instanceAbbr[instanceName] then
        -- Has no abbreviation yet, try it with a somewhat good guess
        -- Tries to get the first char of every word, does not go well with utf-8 chars
        self.db.profile.instanceAbbr[instanceName] = string_gsub(instanceName, "(%a)[%l%p]*[%s%-]*", "%1")
    end

    return ( self.db.profile.instanceAbbr[instanceName] ~= "" and self.db.profile.instanceAbbr[instanceName] or nil )
end


function Ailo:SetupClasscoloredFonts()
    local class, color, CHOOSEN_CLASS_COLORS

    if self.db.profile.useCustomClassColors and CUSTOM_CLASS_COLORS then
        CHOOSEN_CLASS_COLORS = CUSTOM_CLASS_COLORS
    else
        CHOOSEN_CLASS_COLORS = RAID_CLASS_COLORS
    end
    for class,color in pairs(CHOOSEN_CLASS_COLORS) do
        if not RAID_CLASS_COLORS_FONTS[class] then 
            RAID_CLASS_COLORS_FONTS[class] = CreateFont("ClassFont"..class)
            RAID_CLASS_COLORS_FONTS[class]:CopyFontObject(GameTooltipText)
        end
        RAID_CLASS_COLORS_FONTS[class]:SetTextColor(color.r, color.g, color.b)
    end
end


function Ailo:BuildSortedKeyTables()
    local c, r
    wipe(sortRealms)
    sortRealms = {}
    wipe(sortRealmsPlayer)
    sortRealmsPlayer = {}
    for r,_ in pairs(self.db.global.chars) do
        tinsert(sortRealms, r)
        sortRealmsPlayer[r] = {}
        for c,_ in pairs(self.db.global.chars[r]) do
            tinsert(sortRealmsPlayer[r],c)
        end
        tsort(sortRealmsPlayer[r])
    end
    tsort(sortRealms)
end


-- Seasonal Quest Helper functions

function Seasonal:GetCurrentEvent()
    local now = Ailo:GetSyncTime()

    for k, event in pairs(self.Events) do
        if ( event.first <= now and now <= event.last) then
            return event
        end
    end

    return nil
end

function Seasonal:LevelAdjustment()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.level or 0
end

function Seasonal:Show()
    return Ailo.db.profile.showSeasonal and self:GetCurrentEvent() ~= nil
end

function Seasonal:Icon()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.icon or nil
end

function Seasonal:Dungeon()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.dungeon_id or nil
end

function Seasonal:Quests()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.quest_ids or nil
end

--=========================================================================--
--=========================================================================--
--
-- Tooltip and Options Display
--
--=========================================================================--
--=========================================================================--

--=========================================================================--
-- PrepareTooltip - build and formate the tooltip to show lockout state
--=========================================================================--

function Ailo:PrepareTooltip(tooltip)
    local raidorder = {}
    --[[[ Cell are just colored green/red
          Foo BaR Etc EtC ETC
    Char1 [ ] [ ] [ ] [ ] [ ]
    Char2 [ ] [ ] [ ] [ ] [ ]
    Char3 [ ] [ ] [ ] [ ] [ ]
    ]]--
    local charsdb = self.db.global.chars
    local raidsdb = self.db.global.raids
    local lfrsdb  = self.db.global.lfrs
    local profile = self.db.profile

    if self:GetSyncTime() > self.db.global.nextPurge then
        -- Only search 
        self:PurgeExpiredData()
        self:TrimRaidTable()
        self.db.global.nextPurge = self:GetNextPurge()
    end

    if next(raidsdb) or next(lfrsdb) or profile.showAllChars or 
       Seasonal:Show() or profile.showWeeklyValor or 
       profile.showDailyHeroic or profile.showDailyScenario or
       profile.showRaidFinder or 
       profile.showWGVictory or profile.showTBVictory or 
       profile.showWeeklyConquest or profile.showDailyPVP then
        -- At least one char is saved to some 
        tooltip:AddHeader(L["Raid"]) -- Raid

        local seasonalColumn, weeklyValorColumn, dailyHeroicColumn, dailyScenarioColumn, weeklyConquestColumn, dailyPVPColumn, wgVictoryColumn, tbVictoryColumn, shaofangerColumn
        local raidfinder_cols = {}

        -- Seasonal column
        if Seasonal:Show() then
            seasonalColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, seasonalColumn, format("|T%s:0|t", Seasonal:Icon()))
        end

        -- Weekly Valor column
        if profile.showWeeklyValor then
            weeklyValorColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, weeklyValorColumn, format("|T%s:0|t", icons.Valor))
        end

        -- Daily Heroic column
        if profile.showDailyHeroic then
            dailyHeroicColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, dailyHeroicColumn, format("|T%s:0|t", icons.LFD))
        end

        -- Daily Scenario column
        if profile.showDailyScenario then
            dailyScenarioColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, dailyScenarioColumn, format("|T%s:0|t", icons.LFS))
        end

        -- Weekly Conquest column
        if profile.showWeeklyConquest then
            weeklyConquestColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, weeklyConquestColumn, format("|T%s:0|t", icons.Conquest))
        end

        -- PvP Daily column
        if profile.showDailyPVP then
            dailyPVPColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, dailyPVPColumn, format("|T%s:0|t", icons.RandomBG))
        end

        -- Wintergrasp Victory column
        if profile.showWGVictory then
            wgVictoryColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, wgVictoryColumn, format("|T%s:0|t", icons.TolBarad))
        end
        -- Tol Barad Victory column
        if profile.showTBVictory then
            tbVictoryColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, tbVictoryColumn, format("|T%s:0|t", icons.Wintergrasp))
        end

        -- Sha of Anger Column
        if profile.showShaOfAnger then
            local name = EJ_GetEncounterInfo(691)
            local abbr = format("|T%s:0|t%s", icons.ShaOfAnger, self:GetInstanceAbbr(name))
            shaofangerColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, shaofangerColumn, abbr)
        end

        -- Raid Finder columns
        if profile.showRaidFinder then
            for _,rf in ipairs(RaidFinder) do
                if lfrsdb[rf.id] then
                    local name = GetLFGDungeonInfo(rf.id)
                    local abbr = format("|T%s:0|t%s", icons.LFR, self:GetInstanceAbbr(name))
                    local column = tooltip:AddColumn("CENTER")
                    raidfinder_cols[rf] = column
                    tooltip:SetCell(1, column, abbr)
                end
            end
        end

        -- Instances with lockouts
        for raid, sizes in pairs(raidsdb) do
            local raidabbr = self:GetInstanceAbbr(raid)
            for size, _ in pairs(sizes) do
                if raidabbr and ( size > 5 or self.db.profile.show5Man ) then
                    local lastcolumn = tooltip:AddColumn("CENTER")
                    raidorder[(string_format("%s.%d", raid, size))] = lastcolumn
                    tooltip:SetCell(1, lastcolumn , raidabbr)
                end
            end
        end
        self:BuildSortedKeyTables()
        local lastline
        for _, realm in ipairs(sortRealms) do
            if profile.showRealmHeaderLines then
                lastline = tooltip:AddLine("")
                tooltip:SetCell(lastline, 1, realm, nil, "CENTER", tooltip:GetColumnCount())
            end
            for _, char in ipairs(sortRealmsPlayer[realm]) do
                local charData = charsdb[realm][char]
                if profile.showAllChars or next(charData.lockouts) or next(charData.resets) then

                    local charRealm = char.." - "..realm
                    local lineName = profile.showCharacterRealm and charRealm or char
                    local lineFont = profile.useClassColors and RAID_CLASS_COLORS_FONTS[ self.db.global.charClass[charRealm] ] or nil
                    lastline = tooltip:AddLine("")
                    tooltip:SetCell(lastline, 1, lineName, lineFont)

                    -- init cells
                    for i = tooltip:GetColumnCount(),2,-1 do
                        tooltip:SetCell(lastline, i, "")
                        tooltip:SetCellColor(lastline, i, profile.freeraid.r, profile.freeraid.g, profile.freeraid.b, profile.freeraid.a)
                    end

                    if seasonalColumn and charData.resets.seasonal then
                        tooltip:SetCellColor(lastline, seasonalColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end

                    if weeklyValorColumn and charData.resets.weeklyvalor then 
                        if charData.resets.weeklyvalor.done then
                            tooltip:SetCellColor(lastline, weeklyValorColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                        elseif charData.resets.weeklyvalor.encounter > 0 then
                            tooltip:SetCell(lastline, weeklyValorColumn, charData.resets.weeklyvalor.encounter)
                        end
                    end

                    if dailyHeroicColumn and charData.resets.dailyheroic then 
                        tooltip:SetCellColor(lastline, dailyHeroicColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end

                    if dailyScenarioColumn and charData.resets.dailyscenario then 
                        tooltip:SetCellColor(lastline, dailyScenarioColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end

                    if shaofangerColumn and charData.resets.shaofanger then
                        tooltip:SetCellColor(lastline, shaofangerColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end

                    for rf, column in pairs(raidfinder_cols) do
                        local id = rf.id
                        if charData.resets[id] then
                            if charData.resets[id].encounter == 0 then
                                tooltip:SetCellColor(lastline, column, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                            elseif profile.showLeft then
                                local strCell = string_format("%s%i", charData.resets[id].done and "" or format("|T%s:0|t ", icons.Valor), charData.resets[id].encounter)
                                tooltip:SetCell(lastline, column, strCell)
                            end
                        end
                    end
                    
                    if weeklyConquestColumn and charData.resets.weeklyconquest then 
                        if charData.resets.weeklyconquest.done then
                            tooltip:SetCellColor(lastline, weeklyConquestColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                        elseif charData.resets.weeklyconquest.encounter > 0 then
                            tooltip:SetCell(lastline, weeklyConquestColumn, charData.resets.weeklyconquest.encounter)
                        end
                    end

                    if dailyPVPColumn and charData.resets.dailypvp then
                        tooltip:SetCellColor(lastline, dailyPVPColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    if wgVictoryColumn and charData.resets.wgvictory then
                        tooltip:SetCellColor(lastline, wgVictoryColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    if tbVictoryColumn and charData.resets.tbvictory then
                        tooltip:SetCellColor(lastline, tbVictoryColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    for instance, instanceData in pairs(charData.lockouts) do
                        if raidorder[instance] then
                            tooltip:SetCellColor(lastline, raidorder[instance], profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                            if instanceData.encounter > 0 and profile.showLeft then
                                tooltip:SetCell(lastline, raidorder[instance], instanceData.encounter)
                            end
                        end
                    end
                end
            end
        end
    else
        -- No saved raids at all
        tooltip:AddHeader(L["No saved raids found"])
    end
end


--=========================================================================--
-- GenerateOptions - build options UI table
--=========================================================================--

function Ailo.GenerateOptions()
    if not Ailo.options then
        local function newCounter()
            local i = 0
            return function()
                    i = i + 1
                    return i
                end
        end
        local count = newCounter()

        Ailo.options = {
            name = "Ailo",
            type = 'group',
            args = {
                genconfig = {
                    name = L["General Settings"],
                    type = 'group',
                    order = 1,
                    get = function(info) return Ailo.db.profile[info[#info]] end,
                    set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                    args = {
                        savedraid = {
                            name = L["Saved raid color"],
                            desc = L["SAVED_RAID_DESC"],
                            type = 'color',
                            order = count(),
                            get  = getColor,
                            set  = setColor,
                            hasAlpha = true,
                        },
                        freeraid = {
                            name = L["Free raid color"],
                            desc = L["FREE_RAID_DESC"],
                            type = 'color',
                            order = count(),
                            get  = getColor,
                            set  = setColor,
                            hasAlpha = true,
                        },
                        useClassColors = {
                            type = "toggle",
                            order = count(),
                            name = L["Color names by class"],
                        },
                        useCustomClassColors = {
                            type = "toggle",
                            order = count(),
                            name = L["Use !ClassColors"],
                            desc = L["Use !ClassColors addon for class colors used to color the names in the tooltip"],
                            get = function(info) return Ailo.db.profile[info[#info]] end,
                            set = function(info, value) 
                                Ailo.db.profile[info[#info]] = value 
                                Ailo:SetupClasscoloredFonts()
                            end,
                            disabled = function() return not Ailo.db.profile.useClassColors or not CUSTOM_CLASS_COLORS end,
                        },
                        showRealmHeaderLines  = {
                            type = "toggle",
                            order = count(),
                            name = L["Show Realm Headers"],
                            desc = L["SHOW_REALMLINES_DESC"],
                        },
                        showCharacterRealm = {
                            name = L["Show character realms"],
                            type = "toggle",
                            order = count(),
                        },
                        minimapIcon = {
                            type = "toggle",
                            name = L["Show minimap button"],
                            desc = L["Show the Ailo minimap button"],
                            order = count(),
                            get = function(info) return not Ailo.db.profile.minimapIcon.hide end,
                            set = function(info, value)
                                Ailo.db.profile.minimapIcon.hide = not value
                                if value then LDBIcon:Show("Ailo") else LDBIcon:Hide("Ailo") end
                            end,
                        },
                        showMessages = {
                            type = "toggle",
                            order = count(),
                            name = L["Chatframe Messages"],
                        },
                        hc = {
                            type = "input",
                            order = count(),
                            disabled = true,
                            hidden = true,
                            name = L["Tooltip abbreviation used for heroic raids"],
                            width = "double",
                        },
                        nhc = {
                            type = "input",
                            order = count(),
                            disabled = true,
                            hidden = true,
                            name = L["Tooltip abbreviation used for nonheroic raids"],
                            width = "double",
                        },
                        weeklyResetDay = {
                            type = "select",
                            order = count(),
                            name = L["Weekly reset day"],
                            desc = L["Day when weekly caps and lockouts are reset"],
                            values = function() return { CalendarGetWeekdayNames() } end,
                            get = function(info) return Ailo.db.global.weeklyResetDay end,
                            set = function(info, value) Ailo.db.global.weeklyResetDay = value end,
                        },
                        wipeDB = {
                            type = "execute",
                            name = L["Wipe Database"],
                            order = count(),
                            confirm = true,
                            func = function() 
                                    Ailo:WipeDB() 
                                    Ailo:UpdatePlayer()
                                end,
                        },
                        showAllChars  = {
                            type = "toggle",
                            order = count(),
                            name = L["Show all chars"],
                            desc = L["Regardles of any saved instances"],
                        },                    
                        showLeft = {
                            type = "toggle",
                            order = count(),
                            name = L["Encounters Left"],
                            desc = L["SHOW_ENCOUTERS_LEFT_DESC"],
                        },


                        headerTrackers = {
                            type = "header",
                            order = count(),
                            name = "",
                        },


                        showWeeklyValor = {
                            type = "toggle",
                            order = count(),
                            image = icons.Valor,
                            name = L["Track Valor Points"],
                            desc = L["TRACK_VALOR_POINTS_DESC"],
                        },
                        showDailyHeroic = {
                            type = "toggle",
                            order = count(),
                            image = icons.LFD,
                            name = L["Track 'Daily Heroic'"],
                            desc = L["TRACK_DAILY_HEROIC_DESC"],
                        },
                        showDailyScenario = {
                            type = "toggle",
                            order = count(),
                            image = icons.LFS,
                            name = L["Track 'Daily Scenario'"],
                            desc = L["TRACK_DAILY_SCENARIO_DESC"],
                        },
                        showSeasonal  = {
                            type = "toggle",
                            order = count(),
                            image = Seasonal:Icon(),
                            name = L["Track Seasonal bosses"],
                            desc = L["If the character has received the reward for a seasonal boss"],
                        },
                        showRaidFinder = {
                            type = "toggle",
                            order = count(),
                            image = icons.LFR,
                            name = L["Track Raid Finder instances"],
                            desc = L["Tracks any raid done through the Raid Finder tool"],
                        },
                        showShaOfAnger = {
                            type = "toggle",
                            order = count(),
                            image = icons.ShaOfAnger,
                            name = L["Track Sha of Anger"],
                            desc = L["If the character has collected loot from Sha of Anger this week"],
                        },
                        show5Man = {
                            type = "toggle",
                            order = count(),
                            name = L["Show 5-man instances"],
                        },
                        Padding_PVEPVPSeparator = {
                            order = count(),
                            type = "description",
                            name = "",
                        },

                        showWeeklyConquest = {
                            type = "toggle",
                            order = count(),
                            image = icons.Conquest,
                            name = L["Track Conquest Points"],
                            desc = L["TRACK_CONQUEST_POINTS_DESC"],
                        },
                        showDailyPVP  = {
                            type = "toggle",
                            order = count(),
                            image = icons.RandomBG,
                            name = L["Track PvP daily"],
                            desc = L["TRACK_DAILY_PVP_DESC"],
                        },

                        showWGVictory  = {
                            type = "toggle",
                            order = count(),
                            image = icons.Wintergrasp,
                            name = L["Track 'WG Victory'"],
                            desc = L["If the character has done the 'Victory in Wintergrasp' weekly pvp quest"],
                        },
                        showTBVictory  = {
                            type = "toggle",
                            order = count(),
                            image = icons.TolBarad,
                            name = L["Track 'TB Victory'"],
                            desc = L["If the character has done the 'Victory in Tol Barad' weekly pvp quest"],
                        },
                    },
                },
                instanceAbbr = { 
                    type = 'group',
                    name = L["Instance Abbreviations"],
                    get = function(info) return Ailo.db.profile.instanceAbbr[info[#info]] end,
                    set = function(info, value) Ailo.db.profile.instanceAbbr[info[#info]] = value end,
                    args = {
                        header = {
                            type = "header",
                            order = 1,
                            name = L["Change the abbreviations used in the tooltip"]
                        },
                    },
                },
            },
        }
        local instance, abbr
        for instance, abbr in pairs(Ailo.db.profile.instanceAbbr) do
            Ailo.options.args.instanceAbbr.args[instance] = {
                type = "input",
                name = instance,
            }
        end
        Ailo.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Ailo.db)
    end

    return Ailo.options
end
